#include <iostream>
#include <fstream>
#include <iomanip>
#include <random>
#include <cmath>

using namespace std;

#define LL long long
#define RNGd(rng) generate_canonical<double,52>(rng)
#define PDD pair<double, double>

/*
 generate_canonical<>
-            0.0  1.0.
 generate_canonical<double,50> -     0  1,  double.
             50  ,    0.
 */

int DIM = 10;

double pi,
    R,   // radii
    Rs,  // starting value, = -R +d/2
    d,   // step
    V_ideal;

LL n = 3;     // <<==    
LL total;

random_device r;             //   ""  
mt19937_64  used_RNG(r());   //  ,     .

/*
 LIST OF RNG ENGINES in C++ 11
 minstd_rand0, minstd_rand, mt19937, mt19937_64,
 ranlux24_base, ranlux48_base, ranlux24, ranlux48,
 knuth_b, default_random_engine
*/

//=======================
PDD Regular() {
    LL inside = 0;

    double x21,x22,x23,x24,x25,x26,x27,x28,x29;
    double a, R2=R*R;
    double x1,x2,x3,x4,x5,x6,x7,x8,x9,x10;
    for(x1 = Rs; x1<R; x1+=d) {
        x21 = x1*x1;
     for(x2 = Rs; x2<R; x2+=d) {
        x22 = x2*x2;
      for(x3 = Rs; x3<R; x3+=d) {
        x23 = x3*x3;
       for(x4 = Rs; x4<R; x4+=d) {
        x24 = x4*x4;
        for(x5 = Rs; x5<R; x5+=d) {
            x25 = x5*x5;
         for(x6 = Rs; x6<R; x6+=d) {
            x26 = x6*x6;
          for(x7 = Rs; x7<R; x7+=d) {
            x27 = x7*x7;
           for(x8 = Rs; x8<R; x8+=d) {
            x28 = x8*x8;
            for(x9 = Rs; x9<R; x9+=d) {
                x29 = x9*x9;
             for(x10 = Rs; x10<R; x10+=d) {
                 //double a = x1*x1+x2*x2+x3*x3+x4*x4+x5*x5+x6*x6+x7*x7 +x8*x8+x9*x9+x10*x10;
                 a = x21+x22+x23+x24+x25+x26+x27+x28+x29 +x10*x10;
                 if (a<=R2)
                    inside +=1;
    }}}}}}}}}}
    double p1      = (double)inside/total;
    double V_eval1 = pow(2*R, DIM) *p1;
    double error1  = abs(V_ideal-V_eval1)/V_ideal;
    return make_pair(V_eval1, error1);
}
//=======================

//=======================
// Monte-Carlo
//
//  x = RNGd(used_RNG);  //  -     .
//                      //    -   (define)
//                     //  " "    :
//  s1 = generate_canonical<double,50>(used_RNG);
//
PDD MonteCarlo() {
    LL inside = 0;
    double a,b, R2=R*R;
    for (LL k=0; k<total; ++k) {
        a = 0;
        for (int m=0; m<DIM; ++m) {
            b = (RNGd(used_RNG) *2 -1) *R;
            a+=b*b;
        }
        if (a<=R2)
            inside +=1;
    }
    double p2 = (double)inside/total;
    double V_eval2  = pow(2*R, DIM) *p2;
    double error2  = abs(V_ideal-V_eval2)/V_ideal;
    return make_pair(V_eval2, error2);
}
//=======================
void Compute(LL use_n=2) {
    cout <<"\n====\n";
    n = use_n;

    total = (LL) pow(n, DIM);
    cout <<"N_points = " <<n <<"^"<<DIM <<" = " <<total <<"\n";

    // regular grid
    //
    d = 2*R/n;
    Rs = -R+d/2;

    cout << "Regular:\n";

    pair<double, double> res1 = Regular();
    cout <<"Volume            Error,%\n";
    cout <<res1.first <<"  " <<res1.second*100 <<"\n";

    // Monte-Carlo
    cout <<"Monte-Carlo:\n";
    cout <<"Volume            Error,%\n";

    double average = 0, least = 1e10;
    for(int i=0; i<10; ++i) {
        PDD res2 = MonteCarlo();
        cout <<res2.first <<"  " <<res2.second*100 <<"\n";
        average += res2.second;
        least = min(least, res2.second);
    }
    average/= 10;
    cout <<"average% = " <<average*100 <<"\nleast%   = " <<least*100 <<"\n\n";
}

//=======================

int main() {
    pi = 2*acos(0.0);   //              PI  C++

// hyper-volume of a 10-sphere of radii = 1.0
//  V = 1/120 *pi^5

    R = 1.0;
    V_ideal = 1.0/120.0 *pow(pi, 5) * pow(R, DIM);
    cout.precision(16);
    cout <<DIM <<"D-sphere\nPrecise vol = " <<V_ideal <<"\n";

//    Compute(2);
//    Compute(3);
//    Compute(4);
//    Compute(5);
    Compute(6);
    Compute(7);
}

/*
##== run-1 ==
10D-sphere
Precise vol = 2.550164039877345

====
N_points = 2^10 = 1024
Regular:
Volume            Error,%
0  100
Monte-Carlo:
Volume            Error,%
2  21.57367256671874
6  135.2789822998438
2  21.57367256671874
1  60.78683628335937
3  17.63949114992189
6  135.2789822998438
3  17.63949114992189
4  56.85265486656252
4  56.85265486656252
4  56.85265486656252
average% = 58.03290929160158
least%   = 17.63949114992189


====
N_points = 3^10 = 59049
Regular:
Volume            Error,%
3.485647513082355  36.68326658899967
Monte-Carlo:
Volume            Error,%
2.375789598469068  6.837773518940519
2.219715829226574  12.95791978411962
2.410472658300733  5.47774101556739
2.32376500872157  8.877822274000215
2.514521837795729  1.397643505448
2.687937136954055  5.402519011417649
2.774644786533218  8.802600269850473
2.687937136954055  5.402519011417649
2.462497248048231  3.437692260507695
2.341106538637403  8.19780602231365
average% = 6.679203667358285
least%   = 1.397643505448


====
N_points = 4^10 = 1048576
Regular:
Volume            Error,%
1  60.78683628335937
Monte-Carlo:
Volume            Error,%
2.4931640625  2.235149444742645
2.4755859375  2.924443338199219
2.6572265625  4.198260227518709
2.544921875  0.205561869564956
2.599609375  1.938908021188828
2.6171875  2.628201914645402
2.6064453125  2.206966757533052
2.58203125  1.249614127732255
2.5166015625  1.31609092013388
2.5751953125  0.9815553913880319
average% = 1.988475201264698
least%   = 0.205561869564956


====
N_points = 5^10 = 9765625
Regular:
Volume            Error,%
3.070754816  20.41401133347068
Monte-Carlo:
Volume            Error,%
2.5717374976  0.8459635296124939
2.5381830656  0.4698119058223859
2.533883904  0.6383956334874844
2.5587351552  0.3361005483814724
2.5434259456  0.264221994035687
2.5652363264  0.5910320389969831
2.5300041728  0.7905321682096458
2.5482493952  0.07507927519192541
2.5476202496  0.09975006460633136
2.5418530816  0.3258989675717106
average% = 0.443678612591612
least%   = 0.07507927519192541

Process returned 0 (0x0)   execution time : 16.697 s
Press any key to continue.

10D-sphere
Precise vol = 2.550164039877345

====
N_points = 6^10 = 60466176
Regular:
Volume            Error,%
3.225524564344866  26.48302281370118
Monte-Carlo:
Volume            Error,%
2.547460583583126  0.106011074266
2.558756287151349  0.3369291990493643
2.554031397652797  0.1516513335996048
2.553472539755119  0.1297367473636172
2.551914511676743  0.06864153725116157
2.551931446764551  0.06930561562193588
2.544276787075141  0.2308578079740781
2.55377737133567  0.1416901580377986
2.542854239699233  0.2866403911202522
2.549137157276161  0.04026731555801968
average% = 0.1561731179841833
least%   = 0.04026731555801968


====
N_points = 7^10 = 282475249
Regular:
Volume            Error,%
2.775333905449536  8.829622802736271
Monte-Carlo:
Volume            Error,%
2.550915064420388  0.02945004836153816
2.550190045146221  0.001019748865915714
2.544893779348434  0.2066635889495273
2.547434971904388  0.1070153892174043
2.550668557867171  0.01978374653303663
2.551741586392938  0.06186058978654461
2.5524049790288  0.08787431382502861
2.54574567699558  0.1732579870421923
2.550741059794587  0.02262677648258668
2.553641136891254  0.1363479744650416
average% = 0.08459001635288159
least%   = 0.001019748865915714


Process returned 0 (0x0)   execution time : 525.521 s
Press any key to continue.

====
10D-sphere
Precise vol = 2.550164039877345

====
N_points = 6^10 = 60466176
Regular:
Volume            Error,%
3.225524564344866  26.48302281370118
Monte-Carlo:
Volume            Error,%
2.555318464326238  0.2021212897794799
2.553828176599096  0.1436823931501564
2.54785009060272  0.09073727173787736
2.546054971295026  0.1611295790413823
2.553608020457586  0.1350493743299162
2.552896746769632  0.1071580827568378
2.546156581821877  0.1571451088166494
2.544801774797202  0.2102713784796565
2.547121881826957  0.1192926416817474
2.535233450185439  0.5854756579747002
average% = 0.1912062777748404
least%   = 0.09073727173787736


====
N_points = 7^10 = 282475249
Regular:
Volume            Error,%
2.775333905449536  8.829622802736271
Monte-Carlo:
Volume            Error,%
2.554445908285579  0.1679056069051726
2.548845134392642  0.0517184566984398
2.550193670242592  0.001161900363403665
2.55009941773695  0.002534038571021854
2.546282191258463  0.1522195654154382
2.548725506212405  0.05640945611521133
2.551864839669546  0.06669374070078667
2.550896938938533  0.02873929087415064
2.548863259874496  0.05100769921105229
2.550704808830879  0.02120526150781165
average% = 0.05995950163624889
least%   = 0.001161900363403665


Process returned 0 (0x0)   execution time : 269.842 s
Press any key to continue.

*/
